package io.zfunny.j2dlt.dlt645.facade;

import io.zfunny.j2dlt.dlt645.Dlt645;
import io.zfunny.j2dlt.dlt645.io.AbstractDlt645Transport;
import io.zfunny.j2dlt.dlt645.io.Dlt645TCPTransaction;
import io.zfunny.j2dlt.dlt645.net.TCPMasterConnection;

import java.net.InetAddress;
import java.net.UnknownHostException;

/**
 * Modbus/TCP Master facade.
 *
 * @author Dieter Wimberger
 * @author Steve O'Hara (4NG)
 * @version 2.0 (March 2016)
 */
public class Dlt645TCPMaster extends AbstractDlt645Master {

    private final TCPMasterConnection connection;
    private boolean reconnecting = false;
    private boolean useRtuOverTcp;

    /**
     * Constructs a new master facade instance for communication
     * with a given slave.
     *
     * @param addr an internet address as resolvable IP name or IP number,
     *             specifying the slave to communicate with.
     */
    public Dlt645TCPMaster(String addr) {
        this(addr, Dlt645.DEFAULT_PORT, Dlt645.DEFAULT_TIMEOUT, false, false);
    }

    /**
     * Constructs a new master facade instance for communication
     * with a given slave.
     *
     * @param addr an internet address as resolvable IP name or IP number,
     *             specifying the slave to communicate with.
     * @param useRtuOverTcp True if the RTU protocol should be used over TCP
     */
    public Dlt645TCPMaster(String addr, boolean useRtuOverTcp) {
        this(addr, Dlt645.DEFAULT_PORT, Dlt645.DEFAULT_TIMEOUT, false, useRtuOverTcp);
    }

    /**
     * Constructs a new master facade instance for communication
     * with a given slave.
     *
     * @param addr an internet address as resolvable IP name or IP number,
     *             specifying the slave to communicate with.
     * @param port the port the slave is listening to.
     */
    public Dlt645TCPMaster(String addr, int port) {
        this(addr, port, Dlt645.DEFAULT_TIMEOUT, false, false);
    }

    /**
     * Constructs a new master facade instance for communication
     * with a given slave.
     *
     * @param addr an internet address as resolvable IP name or IP number,
     *             specifying the slave to communicate with.
     * @param port the port the slave is listening to.
     * @param useRtuOverTcp True if the RTU protocol should be used over TCP
     */
    public Dlt645TCPMaster(String addr, int port, boolean useRtuOverTcp) {
        this(addr, port, Dlt645.DEFAULT_TIMEOUT, false, useRtuOverTcp);
    }

    /**
     * Constructs a new master facade instance for communication
     * with a given slave.
     *
     * @param addr      an internet address as resolvable IP name or IP number,
     *                  specifying the slave to communicate with.
     * @param port      the port the slave is listening to.
     * @param timeout   Socket timeout in milliseconds
     * @param reconnect True if the socket should reconnect if it detects a connection failure
     */
    public Dlt645TCPMaster(String addr, int port, int timeout, boolean reconnect) {
        this(addr, port, timeout, reconnect, false);
    }

    /**
     * Constructs a new master facade instance for communication
     * with a given slave.
     *
     * @param addr      an internet address as resolvable IP name or IP number,
     *                  specifying the slave to communicate with.
     * @param port      the port the slave is listening to.
     * @param timeout   Socket timeout in milliseconds
     * @param reconnect True if the socket should reconnect if it detcts a connection failure
     * @param useRtuOverTcp True if the RTU protocol should be used over TCP
     */
    public Dlt645TCPMaster(String addr, int port, int timeout, boolean reconnect, boolean useRtuOverTcp) {
        super();
        this.useRtuOverTcp = useRtuOverTcp;
        try {
            InetAddress slaveAddress = InetAddress.getByName(addr);
            connection = new TCPMasterConnection(slaveAddress);
            connection.setPort(port);
            connection.setTimeout(timeout);
            this.timeout = timeout;
            setReconnecting(reconnect);
        }
        catch (UnknownHostException e) {
            throw new RuntimeException("Failed to contruct Dlt645TCPMaster instance.", e);
        }
    }

    /**
     * Connects this <tt>ModbusTCPMaster</tt> with the slave.
     *
     * @throws Exception if the connection cannot be established.
     */
    @Override
    public synchronized void connect() throws Exception {
        if (connection != null && !connection.isConnected()) {
            connection.connect(useRtuOverTcp);
            transaction = connection.getDlt645Transport().createTransaction();
            ((Dlt645TCPTransaction)transaction).setReconnecting(reconnecting);
            setTransaction(transaction);
        }
    }

    /**
     * Disconnects this <tt>ModbusTCPMaster</tt> from the slave.
     */
    @Override
    public synchronized void disconnect() {
        if (connection != null && connection.isConnected()) {
            connection.close();
            transaction = null;
            setTransaction(null);
        }
    }

    /**
     * Tests if a constant connection is maintained or if a new
     * connection is established for every transaction.
     *
     * @return true if a new connection should be established for each
     * transaction, false otherwise.
     */
    public synchronized boolean isReconnecting() {
        return reconnecting;
    }

    /**
     * Sets the flag that specifies whether to maintain a
     * constant connection or reconnect for every transaction.
     *
     * @param b true if a new connection should be established for each
     *          transaction, false otherwise.
     */
    public synchronized void setReconnecting(boolean b) {
        reconnecting = b;
        if (transaction != null) {
            ((Dlt645TCPTransaction)transaction).setReconnecting(b);
        }
    }

    @Override
    public synchronized void setTimeout(int timeout) {
        super.setTimeout(timeout);
        if (connection != null) {
            connection.setTimeout(timeout);
        }
    }

    @Override
    public AbstractDlt645Transport getTransport() {
        return connection == null ? null : connection.getDlt645Transport();
    }

    @Override
    public boolean isConnected() {
        return connection != null && connection.isConnected();
    }
}
